use anyhow::{Context, Result};

use cpal::Sample;

use tracing::{error, info, trace, warn};

use tokio::sync::mpsc::{Receiver, Sender};
use tokio::sync::mpsc::error::TryRecvError;

use crate::task_msg::{RawEthFrameMessage, PcmSampleMessage, SampleRateNotifyMessage};
use crate::dsp_common::carrier::{Carrier, TwoStateCarrier};

pub async fn dsp_outb_main(
    tx_for_wire: Sender<PcmSampleMessage>,
    mut rx_for_mod: Receiver<RawEthFrameMessage>,
    mut rx_sample_rate_notify: tokio::sync::oneshot::Receiver<SampleRateNotifyMessage>,
) -> Result<()> {
    trace!("pcm_outb task started");

    // TODO: fetch real sample rate
    let mut carrier = TwoStateCarrier::new(
        320, // freq_low
        384, // freq_high
        rx_sample_rate_notify.await?, // sample_rate
        4, // symbol_rate
    );

    loop {
        match rx_for_mod.try_recv() {
            Ok(eth_frame) => {
                trace!("network frame came in");
                let mut txed = 0;
                let mut bytes = 0;

                for byte in eth_frame {
                    for samp in carrier.byte_into_fsk_samples(&byte) {
                        tx_for_wire.send(samp).await?;
                        txed += 1;
                    }
                    bytes += 1;
                }

                trace!("eth frame modulated and sent ({} bytes, {} samps)", bytes, txed);
                //(*carrier_phase * TAU * freq / sample_rate).sin()
            },
            Err(TryRecvError::Empty) => {
                trace!("no eth frames to modulate!");
                //tx_for_wire.send(i64::EQUILIBRIUM).await?;
                //continue;
                for samp in carrier.idle() {
                    tx_for_wire.send(samp).await?;
                }
            }
            Err(TryRecvError::Disconnected) => {
                Err(anyhow::Error::msg(
                    format!("rx_for_mod closed?! the tap task likely crashed"),
                ))?
            }
        };
    }
}
